<h3>Download</h3>
<p>Current version: <b>{{ release }}</b></p>
<p>Get PyTango from <a href="https://pypi.org/project/pytango/">PyPI</a><br/>
or install it with:</p>
<pre>python -m pip install pytango</pre>

<h3>PDF</h3>
<p>A PDF version <a href="https://media.readthedocs.org/pdf/pytango/latest/pytango.pdf">here</a>.


<h3>Other versions</h3>
<a href="https://pytango.readthedocs.io/en/latest">Development</a>
<br/>
<a href="https://pytango.readthedocs.io/en/stable">Latest stable</a>
<br/>
<a href="https://www.esrf.eu/computing/cs/tango/pytango/v818">8.1.8</a>
<br/>
<a href="https://www.esrf.eu/computing/cs/tango/pytango/v816">8.1.6</a>
